"""Join a sequence of arrays along a new axis."""
from __future__ import annotations
from typing import Optional, Sequence

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.stack)
def stack(
    arrays: Sequence[PolyLike],
    axis: int = 0,
    out: Optional[ndpoly] = None,
) -> ndpoly:
    """
    Join a sequence of arrays along a new axis.

    The ``axis`` parameter specifies the index of the new axis in the
    dimensions of the result. For example, if ``axis=0`` it will be the first
    dimension and if ``axis=-1`` it will be the last dimension.

    Args:
        arrays:
            Each array must have the same shape.
        axis:
            The axis in the result array along which the input arrays are
            stacked.
        out:
            If provided, the destination to place the result. The shape must be
            correct, matching that of what stack would have returned if no out
            argument were specified.

    Returns:
        The stacked array has one more dimension than the input arrays.

    Examples:
        >>> poly = numpoly.variable(3)
        >>> const = numpoly.polynomial([1, 2, 3])
        >>> numpoly.stack([poly, const])
        polynomial([[q0, q1, q2],
                    [1, 2, 3]])
        >>> numpoly.stack([poly, const], axis=-1)
        polynomial([[q0, 1],
                    [q1, 2],
                    [q2, 3]])

    """
    arrays = numpoly.align_exponents(*arrays)
    arrays = numpoly.align_dtype(*arrays)
    result = numpy.stack(
        [array.values for array in arrays], axis=axis, out=out)
    return numpoly.aspolynomial(result, names=arrays[0].indeterminants)




"""Multi indices for monomial exponents."""
from __future__ import annotations
from typing import Optional

import numpy
import numpy.typing

from .cross_truncation import cross_truncate
from .glexsort import glexsort


def glexindex(
        start: numpy.typing.ArrayLike,
        stop: Optional[numpy.typing.ArrayLike] = None,
        dimensions: int = 1,
        cross_truncation: numpy.typing.ArrayLike = 1.,
        graded: bool = False,
        reverse: bool = False,
) -> numpy.ndarray:
    """
    Generate graded lexicographical multi-indices for the monomial exponents.

    Args:
        start:
            The lower order of the indices. If array of int, counts as lower
            bound for each axis.
        stop:
            The maximum shape included. If omitted: stop <- start; start <- 0
            If int is provided, set as largest total order. If array of int,
            set as upper bound for each axis.
        dimensions:
            The number of dimensions in the expansion.
        cross_truncation:
            Use hyperbolic cross truncation scheme to reduce the number of
            terms in expansion. If two values are provided, first is low bound
            truncation, while the latter upper bound. If only one value, upper
            bound is assumed.
        graded:
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``(2, 2, 2)`` has a sum of 6, and will therefore be
            consider larger than both ``(3, 1, 1)`` and ``(1, 1, 3)``.
        reverse:
            Reversed lexicographical sorting meaning that ``(1, 3)`` is
            considered smaller than ``(3, 1)``, instead of the opposite.

    Returns:
        list:
            Order list of indices.

    Examples:
        >>> numpoly.glexindex(4).tolist()
        [[0], [1], [2], [3]]
        >>> numpoly.glexindex(2, dimensions=2).tolist()
        [[0, 0], [1, 0], [0, 1]]
        >>> numpoly.glexindex(start=2, stop=3, dimensions=2).tolist()
        [[2, 0], [1, 1], [0, 2]]
        >>> numpoly.glexindex([1, 2, 3]).tolist()
        [[0, 0, 0], [0, 1, 0], [0, 0, 1], [0, 0, 2]]
        >>> numpoly.glexindex([1, 2, 3], cross_truncation=numpy.inf).tolist()
        [[0, 0, 0], [0, 1, 0], [0, 0, 1], [0, 1, 1], [0, 0, 2], [0, 1, 2]]

    """
    if stop is None:
        start, stop = 0, start
    start = numpy.array(start, dtype=int).flatten()
    stop = numpy.array(stop, dtype=int).flatten()
    start, stop, _ = numpy.broadcast_arrays(start, stop, numpy.empty(dimensions))

    cross_truncation = cross_truncation*numpy.ones(2)
    indices = _glexindex(start, stop, cross_truncation)

    if indices.size:
        indices = indices[glexsort(indices.T, graded=graded, reverse=reverse)]
    return indices


def _glexindex(start, stop, cross_truncation=1.):
    """Backend for the glexindex function."""
    # At the beginning the current list of indices just ranges over the
    # last dimension.
    bound = stop.max()
    dimensions = len(start)
    start = numpy.clip(start, a_min=0, a_max=None)
    dtype = numpy.uint8 if bound < 256 else numpy.uint16
    range_ = numpy.arange(bound, dtype=dtype)
    indices = range_[:, numpy.newaxis]

    for idx in range(dimensions-1):

        # Truncate at each step to keep memory usage low
        if idx:
            indices = indices[cross_truncate(indices, bound-1, cross_truncation[1])]

        # Repeats the current set of indices.
        # e.g. [0,1,2] -> [0,1,2,0,1,2,...,0,1,2]
        indices = numpy.tile(indices, (bound, 1))

        # Stretches ranges over the new dimension.
        # e.g. [0,1,2] -> [0,0,...,0,1,1,...,1,2,2,...,2]
        front = range_.repeat(len(indices)//bound)[:, numpy.newaxis]

        # Puts them two together.
        indices = numpy.column_stack((front, indices))

    # Complete the truncation scheme
    if dimensions == 1:
        indices = indices[(indices >= start) & (indices < bound)]
    else:
        lower = cross_truncate(indices, start-1, cross_truncation[0])
        upper = cross_truncate(indices, stop-1, cross_truncation[1])
        indices = indices[lower^upper]

    return numpy.array(indices, dtype=int).reshape(-1, dimensions)



"""Save several arrays into a single file in uncompressed ``.npz`` format."""
from __future__ import annotations
from os import PathLike
from typing import Dict

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.savez)
def savez(file: PathLike, *args: PolyLike, **kwargs: PolyLike) -> None:
    """
    Save several arrays into a single file in uncompressed ``.npz`` format.

    If arguments are passed in with no keywords, the corresponding variable
    names, in the ``.npz`` file, are 'arr_0', 'arr_1', etc. If keyword
    arguments are given, the corresponding variable names, in the ``.npz``
    file will match the keyword names.

    Args:
        file:
            Either the filename (string) or an open file (file-like object)
            where the data will be saved. If file is a string or a Path, the
            ``.npz`` extension will be appended to the filename if it is not
            already there.
        args:
            Arrays to save to the file. Since it is not possible for Python to
            know the names of the arrays outside `savez`, the arrays will be
            saved with names "arr_0", "arr_1", and so on. These arguments can
            be any expression.
        kwds:
            Arrays to save to the file. Arrays will be saved in the file with
            the keyword names.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([q0, q1-1])
        >>> array = numpy.array([1, 2])
        >>> numpoly.savez("/tmp/savez.npz", a=array, p=poly)
        >>> numpoly.load("/tmp/savez.npz")
        {'a': array([1, 2]), 'p': polynomial([q0, q1-1])}
        >>> numpoly.savez("/tmp/savez.npz", array, poly)
        >>> out = numpoly.load("/tmp/savez.npz")
        >>> out["arr_0"], out["arr_1"]
        (array([1, 2]), polynomial([q0, q1-1]))

    """
    for idx, arg in enumerate(args):
        assert "arr_%d" % idx not in kwargs, "naming conflict"
        kwargs["arr_%d" % idx] = arg

    polynomials = {
        key: numpoly.aspolynomial(kwargs.pop(key))
        for key, value in list(kwargs.items())
        if isinstance(value, numpoly.ndpoly)
    }
    kwargs.update({"-".join(poly.names)+"-"+key: poly.values
                   for key, poly in polynomials.items()})
    numpy.savez(file, **kwargs)



"""Convert inputs to arrays with at least one dimension."""
from __future__ import annotations
from typing import List, Union

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.atleast_1d)
def atleast_1d(*arys: PolyLike) -> Union[ndpoly, List[ndpoly]]:
    """
    Convert inputs to arrays with at least one dimension.

    Scalar inputs are converted to 1-dimensional arrays, whilst
    higher-dimensional inputs are preserved.

    Args:
        arys:
            One or more input arrays.

    Returns:
        An array, or list of arrays, each with ``a.ndim >= 1``.
        Copies are made only if necessary.

    Examples:
        >>> numpoly.atleast_1d(numpoly.variable())
        polynomial([q0])
        >>> numpoly.atleast_1d(1, [2, 3])
        [polynomial([1]), polynomial([2, 3])]

    """
    if len(arys) == 1:
        poly = numpoly.aspolynomial(arys[0])
        array = numpy.atleast_1d(poly.values)
        return numpoly.aspolynomial(array, names=poly.indeterminants)
    return [atleast_1d(ary) for ary in arys]



# pylint: disable=wildcard-import
"""Numpoly -- Multivariate polynomials as numpy elements."""
import logging
import os
import pkg_resources

from .baseclass import ndpoly, FeatureNotSupported, PolyLike

from .align import (
    align_polynomials,
    align_exponents,
    align_indeterminants,
    align_shape,
    align_dtype,
)
from .construct import (
    polynomial,
    aspolynomial,
    clean_attributes,
    remove_redundant_coefficients,
    remove_redundant_names,
    polynomial_from_attributes,
    polynomial_from_roots,
    monomial,
    symbols,
    variable,
)
from .sympy_ import to_sympy

from .array_function import *
from .poly_function import *
from .utils import (
    bindex,
    cross_truncate,
    glexindex,
    glexsort,
)
from .option import get_options, set_options, global_options
from .dispatch import FUNCTION_COLLECTION, UFUNC_COLLECTION


def get_version(name: str) -> str:
    """
    Get distribution version number, if it exists.

    Examples:
        >>> get_version("numpy") == ""
        False
        >>> get_version("not_an_distribution") == ""
        True

    """
    version = ""
    try:
        version = pkg_resources.get_distribution(name).version
    except pkg_resources.DistributionNotFound:
        pass
    return version


def configure_logging() -> None:
    """Configure logging for Numpoly."""
    logpath = os.environ.get("NUMPOLY_LOGPATH", os.devnull)
    logging.basicConfig(level=logging.DEBUG, filename=logpath, filemode="w")
    streamer = logging.StreamHandler()
    loglevel = logging.DEBUG if os.environ.get("NUMPOLY_DEBUG", "") else logging.WARNING
    streamer.setLevel(loglevel)

    logger = logging.getLogger(__name__)
    logger.addHandler(streamer)


__version__ = get_version("numpoly")
configure_logging()



"""Truncation rules for indices."""
from __future__ import annotations

import numpy
import numpy.typing


def cross_truncate(
        indices: numpy.typing.ArrayLike,
        bound: numpy.typing.ArrayLike,
        norm: float,
) -> numpy.ndarray:
    r"""
    Truncate of indices using L_p norm.

    .. math:
        L_p(x) = (\sum_i |x_i/b_i|^p )^{1/p} \leq 1

    where :math:`b_i` are bounds that each :math:`x_i` should follow.

    Args:
        indices:
            Indices to be truncated.
        bound:
            The bound function for witch the indices can not be larger than.
        norm:
            The `p` in the `L_p`-norm. Support includes both `L_0` and `L_inf`.

    Returns:
        Boolean indices to ``indices`` with True for each index where the
        truncation criteria holds.

    Examples:
        >>> indices = numpy.array(numpy.mgrid[:10, :10]).reshape(2, -1).T
        >>> indices[cross_truncate(indices, 2, norm=0.)].T
        array([[0, 0, 0, 1, 2],
               [0, 1, 2, 0, 0]])
        >>> indices[cross_truncate(indices, 2, norm=1.)].T
        array([[0, 0, 0, 1, 1, 2],
               [0, 1, 2, 0, 1, 0]])
        >>> indices[cross_truncate(indices, [0, 1], norm=1.)].T
        array([[0, 0],
               [0, 1]])

    """
    assert norm >= 0, "negative L_p norm not allowed"
    indices = numpy.asarray(indices)
    bound_ = numpy.broadcast_to(numpy.asfarray(bound).ravel(), (indices.shape[1],))
    nudge_factor = 1e-12*indices.shape[1]

    if numpy.any(bound_ < 0):
        return numpy.zeros((len(indices),), dtype=bool)

    if numpy.any(bound_ == 0):
        out = numpy.all(indices[:, bound_ == 0] == 0, axis=-1)
        if numpy.any(bound_):
            out &= cross_truncate(indices[:, bound_ != 0], bound_[bound_ != 0], norm=norm)
        return out

    if norm == 0:
        out = numpy.sum(indices > 0, axis=-1) <= 1+nudge_factor
        out[numpy.any(indices > bound_, axis=-1)] = False
    elif norm == numpy.inf:
        out = numpy.max(indices/bound_, axis=-1) <= 1+nudge_factor
    else:
        out = numpy.sum((indices/bound_)**norm, axis=-1)**(1./norm) <= 1+nudge_factor

    assert numpy.all(out[numpy.all(indices == 0, axis=-1)])
    return out



"""Move axes of an array to new positions."""
from __future__ import annotations
from typing import Sequence, Union
import numpy

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements, simple_dispatch


@implements(numpy.moveaxis)
def moveaxis(
    a: PolyLike,
    source: Union[int, Sequence[int]],
    destination: Union[int, Sequence[int]],
) -> ndpoly:
    """
    Move axes of an array to new positions.

    Other axes remain in their original order.

    Args:
        a:
            The array whose axes should be reordered.
        source:
            Original positions of the axes to move. These must be unique.
        destination:
            Destination positions for each of the original axes. These must
            also be unique.

    Returns:
        Array with moved axes. This array is a view of the input array.

    Examples:
        >>> poly = numpoly.monomial(6).reshape(1, 2, 3)
        >>> numpoly.moveaxis(poly, 0, -1).shape
        (2, 3, 1)
        >>> numpoly.moveaxis(poly, [0, 2], [2, 0]).shape
        (3, 2, 1)

    """
    return simple_dispatch(
        numpy_func=numpy.moveaxis,
        inputs=(a,),
        source=source,
        destination=destination,
    )



"""Compute the determinant of an polynomial array."""
from __future__ import annotations

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.linalg.det)
def det(a: PolyLike) -> ndpoly:
    """
    Compute the determinant of an polynomial array.

    Args:
        a:
            Input array to compute determinants for.
            Shape on form `(..., M, M)`.

    Returns:
        Determinant of `a`. Shape `(...)`.

    Notes:
        Broadcasting rules apply, see the `numpy.linalg` documentation for
        details.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> a = numpoly.polynomial([[[1, 2], [3, 4]], [[1, q0], [q1, 1]]])
        >>> numpoly.det(a)
        polynomial([-2, -q0*q1+1])

    """
    a = numpoly.aspolynomial(a)
    assert a.ndim >= 2, a
    assert a.shape[-2] == a.shape[-1], a.shape
    dims = a.shape[-1]
    index = (slice(None),)*(a.ndim-2)
    if dims == 2:
        return a[index+(0, 0)]*a[index+(1, 1)]-a[index+(1, 0)]*a[index+(0, 1)]
    out = numpoly.zeros_like(a, shape=a.shape[:-2])
    r = numpy.arange(1, dims, dtype=int)
    for idx in range(dims):
        idx0 = index+(0, idx)
        idx1 = index+(slice(1, None), (r+idx)%dims)
        out = out+a[idx0]*det(a[idx1])
    return out



"""Adjust the dimensions of a polynomial."""
from __future__ import annotations
from typing import Optional

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike


def set_dimensions(poly: PolyLike, dimensions: Optional[int] = None) -> ndpoly:
    """
    Adjust the dimensions of a polynomial.

    Args:
        poly:
            Input polynomial
        dimensions:
            The dimensions of the output polynomial. If omitted, increase
            polynomial with one dimension.

    Returns:
        Polynomials with no internal dimensions. Unless the new dim is smaller
        then `poly`'s dimensions, the polynomial should have the same content.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = q0*q1-q0**2
        >>> numpoly.set_dimensions(poly, 1)
        polynomial(-q0**2)
        >>> numpoly.set_dimensions(poly, 3)
        polynomial(q0*q1-q0**2)
        >>> numpoly.set_dimensions(poly).names
        ('q0', 'q1', 'q2')

    """
    poly = numpoly.aspolynomial(poly)
    if dimensions is None:
        dimensions = len(poly.names)+1
    diff = dimensions-len(poly.names)
    if diff > 0:
        padding = numpy.zeros((len(poly.exponents), diff), dtype="uint32")
        exponents = numpy.hstack([poly.exponents, padding])
        coefficients = poly.coefficients
        varname = numpoly.get_options()["default_varname"]
        names_ = list(poly.names)
        idx = 0
        while len(names_) < dimensions:
            candidate = "%s%d" % (varname, idx)
            if candidate not in names_:
                names_.append(candidate)
            idx += 1

        indices = numpy.lexsort([names_])
        exponents = exponents[:, indices]
        names = tuple(names_[idx] for idx in indices)

    elif diff < 0:
        indices = True ^ numpy.any(poly.exponents[:, dimensions:], -1)
        exponents = poly.exponents[:, :dimensions]
        exponents = exponents[indices]
        coefficients = [
            coeff for coeff, idx in zip(poly.coefficients, indices) if idx]
        names = poly.names[:dimensions]

    else:
        return poly

    return numpoly.polynomial_from_attributes(
        exponents=exponents,
        coefficients=coefficients,
        names=names,
        dtype=poly.dtype,
        allocation=poly.allocation,
        retain_names=True,
    )



"""Expand the shape of an array."""
from __future__ import annotations
import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.expand_dims)
def expand_dims(a: PolyLike, axis: int) -> ndpoly:
    """
    Expand the shape of an array.

    Insert a new axis that will appear at the `axis` position in the expanded
    array shape.

    Args:
        a:
            Input array.
        axis:
            Position in the expanded axes where the new axis is placed.

    Returns:
        View of `a` with the number of dimensions increased by one.

    Examples:
        >>> poly = numpoly.variable(2)
        >>> numpoly.expand_dims(poly, axis=0)
        polynomial([[q0, q1]])
        >>> numpoly.expand_dims(poly, axis=1)
        polynomial([[q0],
                    [q1]])

    """
    a = numpoly.aspolynomial(a)
    out = numpy.expand_dims(a.values, axis=axis)
    return numpoly.polynomial(out, names=a.indeterminants)



"""Return the string representation of an array."""
from __future__ import annotations
from typing import List, Optional, Union

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.array_repr)
def array_repr(
        arr: PolyLike,
        max_line_width: Optional[int] = None,
        precision: Optional[int] = None,
        suppress_small: Optional[bool] = None,
) -> str:
    """
    Return the string representation of an array.

    Args:
        arr:
            Input array.
        max_line_width:
            Inserts newlines if text is longer than `max_line_width`. Defaults
            to ``numpy.get_printoptions()['linewidth']``.
        precision:
            Floating point precision. Defaults to
            ``numpy.get_printoptions()['precision']``.
        suppress_small:
            Represent numbers "very close" to zero as zero; default is False.
            Very close is defined by precision: if the precision is 8, e.g.,
            numbers smaller (in absolute value) than 5e-9 are represented as
            zero. Defaults to ``numpy.get_printoptions()['suppress']``.

    Returns:
        The string representation of an array.

    Examples:
        >>> q0 = numpoly.variable()
        >>> numpoly.array_repr(numpoly.polynomial([1, q0]))
        'polynomial([1, q0])'
        >>> numpoly.array_repr(numpoly.polynomial([]))
        'polynomial([], dtype=int64)'
        >>> numpoly.array_repr(
        ...     numpoly.polynomial([1e-6, 4e-7*q0, 2*q0, 3]),
        ...     precision=4,
        ...     suppress_small=True,
        ... )
        'polynomial([0.0, 0.0, 2.0*q0, 3.0])'

    """
    prefix = "polynomial("
    suffix = ")"
    arr = numpoly.aspolynomial(arr)
    if not arr.size:
        return prefix + "[], dtype=%s" % arr.dtype.name + suffix

    arr = to_string(arr, precision=precision, suppress_small=suppress_small)

    return prefix + numpy.array2string(
        numpy.array(arr),
        max_line_width=max_line_width,
        separator=", ",
        formatter={"all": str},
        prefix=prefix,
        suffix=suffix,
    ) + suffix


def to_string(
        poly: ndpoly,
        precision: Optional[float] = None,
        suppress_small: Optional[bool] = None,
) -> Union[str, List]:
    """
    Convert numpoly object to string object, or array of string objects.

    Args:
        poly:
            Polynomial object to convert to strings.
        precision:
            Floating point precision. Defaults to
            ``numpy.get_printoptions()['precision']``.
        suppress_small:
            Represent numbers "very close" to zero as zero; default is False.
            Very close is defined by precision: if the precision is 8, e.g.,
            numbers smaller (in absolute value) than 5e-9 are represented as
            zero. Defaults to ``numpy.get_printoptions()['suppress']``.

    Returns:
        If scalar, a string, or if array, numpy.array with string values.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([[1, q0**3], [q1-1, -3*q0]])
        >>> string_array = to_string(poly)
        >>> string_array
        [['1', 'q0**3'], ['q1-1', '-3*q0']]
        >>> type(string_array[0][0]) == str
        True

    """
    if precision is None:
        precision = numpy.get_printoptions()["precision"]
    if suppress_small is None:
        suppress_small = numpy.get_printoptions()["suppress"]

    if poly.shape:
        return [to_string(poly_, precision=precision,
                          suppress_small=suppress_small)
                for poly_ in poly]

    output = _to_string(poly, precision, suppress_small)
    if output:
        return "".join(output)
    return str(numpy.zeros(1, dtype=poly.dtype).item())


def _to_string(
        poly: ndpoly,
        precision: float,
        suppress_small: bool,
) -> List[str]:
    """Backend for to_string."""
    exponents = poly.exponents.copy()
    coefficients = poly.coefficients
    options = numpoly.get_options()
    output: List[str] = []
    indices = numpoly.glexsort(
        exponents.T,
        graded=options["display_graded"],
        reverse=options["display_reverse"],
    )
    if options["display_inverse"]:
        indices = indices[::-1]
    for idx in indices:

        if not coefficients[idx]:
            continue
        if suppress_small and abs(coefficients[idx]) < 10**-precision:
            continue

        if coefficients[idx] == 1 and any(exponents[idx]):
            out = ""
        elif coefficients[idx] == -1 and any(exponents[idx]):
            out = "-"
        else:
            out = str(coefficients[idx])

        exps_and_names = list(zip(exponents[idx], poly.names))
        for exponent, indeterminant in exps_and_names:
            if exponent:
                if out not in ("", "-"):
                    out += options["display_multiply"]
                out += indeterminant
            if exponent > 1:
                out += options["display_exponent"]+str(exponent)
        if output and float(coefficients[idx]) >= 0:
            out = "+"+out
        output.append(out)

    return output



"""Return a new array of given shape and type, filled with ones."""
from __future__ import annotations
from typing import Any, Sequence

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly
from ..dispatch import implements

Order = Any
try:
    from typing import Literal, Union
    Order = Union[Literal["C"], Literal["F"], None]  # type: ignore
except ImportError:
    pass


@implements(numpy.ones)
def ones(
    shape: Union[int, Sequence[int]],
    dtype: numpy.typing.DTypeLike = float,
    order: Order = "C",
) -> ndpoly:
    """
    Return a new array of given shape and type, filled with ones.

    Args:
        shape:
            Shape of the new array, e.g., ``(2, 3)`` or ``2``.
        dtype:
            The desired data-type for the array, e.g., `numpy.int8`.
            Default is `numpy.float64`.
        order:
            Whether to store multi-dimensional data in row-major
            (C-style) or column-major (Fortran-style) order in
            memory.

    Returns:
        Array of ones with the given shape, dtype, and order.

    Examples:
        >>> numpoly.ones(5)
        polynomial([1.0, 1.0, 1.0, 1.0, 1.0])

    """
    return numpoly.polynomial(numpy.ones(shape, dtype=dtype, order=order))



"""Sum of array elements over a given axis."""
from __future__ import annotations
from typing import Any, Optional, Sequence, Union

import numpy
import numpy.typing

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements, simple_dispatch


@implements(numpy.sum)
def sum(
    a: PolyLike,
    axis: Union[None, int, Sequence[int]] = None,
    dtype: Optional[numpy.typing.DTypeLike] = None,
    out: Optional[ndpoly] = None,
    keepdims: bool = False,
    **kwargs: Any,
) -> ndpoly:
    """
    Sum of array elements over a given axis.

    Args:
        a:
            Elements to sum.
        axis:
            Axis or axes along which a sum is performed. The default,
            axis=None, will sum all of the elements of the input array. If axis
            is negative it counts from the last to the first axis. If axis is
            a tuple of ints, a sum is performed on all of the axes specified in
            the tuple instead of a single axis or all the axes as before.
        dtype:
            The type of the returned array and of the accumulator in which the
            elements are summed.  The dtype of `a` is used by default unless
            `a` has an integer dtype of less precision than the default
            platform integer.  In that case, if `a` is signed then the platform
            integer is used while if `a` is unsigned then an unsigned integer
            of the same precision as the platform integer is used.
        out:
            Alternative output array in which to place the result. It must have
            the same shape as the expected output, but the type of the output
            values will be cast if necessary.
        keepdims:
            If this is set to True, the axes which are reduced are left
            in the result as dimensions with size one. With this option,
            the result will broadcast correctly against the input array.
        initial:
            Starting value for the sum.
        where:
            Elements to include in the sum.

    Returns:
        An array with the same shape as `a`, with the specified axis removed.
        If `a` is a 0-d array, or if `axis` is None, a scalar is returned. If
        an output array is specified, a reference to `out` is returned.

    Examples:
        >>> q0, q1, q2 = numpoly.variable(3)
        >>> poly = numpoly.polynomial([[1, q0], [q1, q2]])
        >>> numpoly.sum(poly)
        polynomial(q2+q1+q0+1)
        >>> numpoly.sum(poly, axis=0)
        polynomial([q1+1, q2+q0])

    """
    return simple_dispatch(
        numpy_func=numpy.sum,
        inputs=(a,),
        out=None if out is None else (out,),
        axis=axis,
        dtype=dtype,
        keepdims=keepdims,
        **kwargs
    )



"""Compose polynomial from array of arrays of polynomials."""
from __future__ import annotations
from typing import List, Optional, Sequence, Set, Tuple

import numpy
import numpy.typing

import numpoly
from ..baseclass import ndpoly, PolyLike


def compose_polynomial_array(
        arrays: Sequence[PolyLike],
        dtype: Optional[numpy.typing.DTypeLike] = None,
        allocation: Optional[int] = None,
) -> ndpoly:
    """
    Compose polynomial from array of arrays of polynomials.

    Backend for `numpoly.polynomial` when input is undetermined.

    Args:
        arrays:
            Input to be converted to a `numpoly.ndpoly` polynomial type.
        dtype:
            Data type used for the polynomial coefficients.
        allocation:
            The maximum number of polynomial exponents. If omitted, use
            length of exponents for allocation.

    Return:
        Polynomial based on input `arrays`.

    """
    oarrays = numpy.array(arrays, dtype=object)
    shape = oarrays.shape
    if not oarrays.size:
        return numpoly.ndpoly(shape=shape, dtype=dtype)
    if len(oarrays.shape) > 1:
        return numpoly.concatenate([numpoly.expand_dims(
            numpoly.aspolynomial(array, dtype=dtype), axis=0)
                                    for array in arrays], axis=0)

    oarrays = oarrays.flatten()
    indices = numpy.array([isinstance(array, numpoly.ndpoly)
                           for array in oarrays])
    oarrays[indices] = numpoly.align_indeterminants(*oarrays[indices])
    names = oarrays[indices][0] if numpy.any(indices) else None
    oarrays = oarrays.tolist()

    dtypes: List[numpy.typing.DTypeLike] = []
    keys: Set[Tuple[int, ...]] = {(0,)}
    for array in oarrays:
        if isinstance(array, numpoly.ndpoly):
            dtypes.append(array.dtype)
            keys = keys.union({tuple(int(k) for k in key)
                               for key in array.exponents.tolist()})
        elif isinstance(array, (numpy.generic, numpy.ndarray)):
            dtypes.append(array.dtype)
        else:
            dtypes.append(type(array))

    if dtype is None:
        dtype = numpy.find_common_type(dtypes, [])
    length = max(1, max([len(key) for key in keys]))

    collection = {}
    for idx, array in enumerate(oarrays):
        if isinstance(array, numpoly.ndpoly):
            for key, value in zip(array.exponents, array.coefficients):
                key = tuple(key)+(0,)*(length-len(key))
                if key not in collection:
                    collection[key] = numpy.zeros(len(oarrays), dtype=dtype)
                collection[key][idx] = value
        else:
            key = (0,)*length
            if key not in collection:
                collection[key] = numpy.zeros(len(oarrays), dtype=dtype)
            collection[key][idx] = array

    exponents = sorted(collection)
    coefficients = numpy.array([collection[key] for key in exponents])
    coefficients = coefficients.reshape(-1, *shape)

    return numpoly.ndpoly.from_attributes(
        exponents=exponents,
        coefficients=list(coefficients),
        names=names,
        allocation=allocation,
    )



"""Return the largest integer smaller or equal to the division of the inputs."""
from __future__ import annotations
from typing import Any, Optional

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements

DIVIDE_ERROR_MSG = """
Divisor in division is a polynomial.
Polynomial division differs from numerical division;
Use ``numpoly.poly_divide`` to get polynomial division."""


@implements(numpy.floor_divide)
def floor_divide(
    x1: PolyLike,
    x2: PolyLike,
    out: Optional[ndpoly] = None,
    where: Optional[numpy.ndarray] = numpy.array(True),
    **kwargs: Any,
) -> ndpoly:
    """
    Return the largest integer smaller or equal to the division of the inputs.

    It is equivalent to the Python ``//`` operator and pairs with the
    Python ``%`` (`remainder`), function so that ``a = a % b + b * (a // b)``
    up to roundoff.

    Args:
        x1:
            Dividend.
        x2:
            Divisor. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        This is a scalar if both `x1` and `x2` are scalars.

    Raises:
        ValueError:
            If `x2` contains indeterminants, numerical division is no longer
            possible and an error is raised instead. For polynomial
            division see ``numpoly.poly_divide``.

    Examples:
        >>> numpoly.floor_divide([1, 3, 5], 2)
        polynomial([0, 1, 2])
        >>> poly = [1, 2, 4]*numpoly.variable(3)
        >>> poly
        polynomial([q0, 2*q1, 4*q2])
        >>> numpoly.floor_divide(poly, 2.)
        polynomial([0.0, q1, 2.0*q2])
        >>> numpoly.floor_divide(poly, [1, 2, 4])
        polynomial([q0, q1, q2])

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    if not x2.isconstant():
        raise numpoly.FeatureNotSupported(DIVIDE_ERROR_MSG)
    x2 = x2.tonumpy()
    dtype = numpy.common_type(x1, x2)
    if x1.dtype == x2.dtype == "int64":
        dtype = "int64"
    no_output = out is None
    if out is None:
        out = numpoly.ndpoly(
            exponents=x1.exponents,
            shape=x1.shape,
            names=x1.indeterminants,
            dtype=dtype,
        )
    assert isinstance(out, numpoly.ndpoly)
    for key in x1.keys:
        out.values[key] = 0
        numpy.floor_divide(
            x1.values[key], x2, out=out.values[key], where=where, **kwargs)
    if no_output:
        out = numpoly.clean_attributes(out)
    return out



"""Simple constructor to create single variables to create polynomials."""
from typing import Optional

import numpy.typing

import numpoly
from ..baseclass import ndpoly


def variable(
        dimensions: int = 1,
        asarray: bool = False,
        dtype: numpy.typing.DTypeLike = "i8",
        allocation: Optional[int] = None,
) -> ndpoly:
    """
    Construct variables that can be used to construct polynomials.

    Args:
        dimensions:
            Number of dimensions in the array.
        asarray:
            Enforce output as array even in the case where there is only one
            variable.
        dtype:
            The data type of the polynomial coefficients.
        allocation:
            The maximum number of polynomial exponents. If omitted, use
            length of exponents for allocation.

    Returns:
        Polynomial array with unit components in each dimension.

    Examples:
        >>> numpoly.variable()
        polynomial(q0)
        >>> q0, q1, q2 = numpoly.variable(3)
        >>> q1+1
        polynomial(q1+1)
        >>> numpoly.polynomial([q2**3, q1+q2, 1])
        polynomial([q2**3, q2+q1, 1])

    """
    return numpoly.symbols(
        names="%s:%d" % (numpoly.get_options()["default_varname"], dimensions),
        asarray=asarray,
        dtype=dtype,
        allocation=allocation,
    )



"""Join a sequence of arrays along an existing axis."""
from __future__ import annotations
from typing import Optional, Sequence

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.concatenate)
def concatenate(
    arrays: Sequence[PolyLike],
    axis: int = 0,
    out: Optional[ndpoly] = None,
) -> ndpoly:
    """
    Join a sequence of arrays along an existing axis.

    Args:
        arrays:
            The arrays must have the same shape, except in the dimension
            corresponding to `axis` (the first, by default).
        axis:
            The axis along which the arrays will be joined.  If axis is None,
            arrays are flattened before use.  Default is 0.
        out:
            If provided, the destination to place the result. The shape must be
            correct, matching that of what concatenate would have returned if
            no out argument were specified.

    Returns:
        The concatenated array.

    Examples:
        >>> const = numpy.array([[1, 2], [3, 4]])
        >>> poly = numpoly.variable(2).reshape(1, 2)
        >>> numpoly.concatenate((const, poly), axis=0)
        polynomial([[1, 2],
                    [3, 4],
                    [q0, q1]])
        >>> numpoly.concatenate((const, poly.T), axis=1)
        polynomial([[1, 2, q0],
                    [3, 4, q1]])
        >>> numpoly.concatenate((const, poly), axis=None)
        polynomial([1, 2, 3, 4, q0, q1])

    """
    arrays = numpoly.align_exponents(*arrays)
    arrays = numpoly.align_dtype(*arrays)
    result = numpy.concatenate(
        [array.values for array in arrays], axis=axis, out=out)
    return numpoly.aspolynomial(result, names=arrays[0].indeterminants)



"""Find the lead coefficients for each polynomial."""
from __future__ import annotations

import numpy
import numpoly

from ..baseclass import PolyLike


def lead_coefficient(
        poly: PolyLike,
        graded: bool = False,
        reverse: bool = False,
) -> numpy.ndarray:
    """
    Find the lead coefficients for each polynomial.

    As polynomials are not inherently sortable, values are sorted using the
    highest `lexicographical` ordering. Between the values that have the same
    highest ordering, the elements are sorted using the coefficients.

    Args:
        poly:
            Polynomial to locate coefficients on.
        graded:
            Graded sorting, meaning the indices are always sorted by the index
            sum. E.g. ``q0**2*q1**2*q2**2`` has an exponent sum of 6, and will
            therefore be consider larger than both ``q0**3*q1*q2``,
            ``q0*q1**3*q2`` and ``q0*q1*z**3``.
        reverse:
            Reverses lexicographical sorting meaning that ``q0*q1**3`` is
            considered bigger than ``q0**3*q1``, instead of the opposite.

    Returns:
        Array of same shape and type as `poly`, containing all the lead
        coefficients.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.lead_coefficient(q0+2*q0**2+3*q0**3)
        3
        >>> numpoly.lead_coefficient([1-4*q1+q0, 2*q0**2-q1, 4])
        array([-4, -1,  4])

    """
    poly = numpoly.aspolynomial(poly)
    out = numpy.zeros(poly.shape, dtype=poly.dtype)
    if not out.size:
        return out
    for idx in numpoly.glexsort(
            poly.exponents.T, graded=graded, reverse=reverse):
        values = poly.coefficients[idx]
        indices = values != 0
        out[indices] = values[indices]
    if not poly.shape:
        out = out.item()
    return out



"""Returns the type that results from applying type promotion rules."""
from __future__ import annotations
from typing import Union

import numpy
import numpy.typing
import numpoly

from ..baseclass import PolyLike
from ..dispatch import implements


@implements(numpy.result_type)
def result_type(
    *arrays_and_dtypes: Union[PolyLike, numpy.typing.DTypeLike],
) -> numpy.dtype:
    """
    Return the type that results from applying type promotion rules.

    Type promotion in NumPy works similarly to the rules in languages like C++,
    with some slight differences.  When both scalars and arrays are used, the
    array's type takes precedence and the actual value of the scalar is taken
    into account.

    For example, calculating 3*a, where a is an array of 32-bit floats,
    intuitively should result in a 32-bit float output.  If the 3 is a 32-bit
    integer, the NumPy rules indicate it can't convert losslessly into a 32-bit
    float, so a 64-bit float should be the result type. By examining the value
    of the constant, '3', we see that it fits in an 8-bit integer, which can be
    cast losslessly into the 32-bit float.

    Args:
        arrays_and_dtypes:
            The operands of some operation whose result type is needed.

    Returns:
        The result type.

    Examples:
        >>> q0 = numpoly.variable(dtype="i1")
        >>> numpoly.result_type(3, numpy.arange(7, dtype="i1"), q0)
        dtype('int8')
        >>> numpoly.result_type('i4', 'c8')
        dtype('complex128')
        >>> numpoly.result_type(3.0, -2)
        dtype('float64')

    """
    values = list(arrays_and_dtypes)
    for idx, value in enumerate(values):
        if isinstance(value, numpoly.ndpoly):
            values[idx] = value.dtype
    return numpy.result_type(*values)



"""Implementation wrapper."""
from typing import Any, Callable, Optional, Sequence, Tuple

import numpoly

from .baseclass import ndpoly

FUNCTION_COLLECTION = {}
UFUNC_COLLECTION = {}


def implements_function(*array_functions: Callable) -> Callable:
    """Register __array_function__."""
    def decorator(numpoly_function: Callable) -> Callable:
        """Register function."""
        for func in array_functions:
            assert func not in FUNCTION_COLLECTION, (
                f"{func} already implemented")
            FUNCTION_COLLECTION[func] = numpoly_function
        return numpoly_function
    return decorator


def implements_ufunc(*array_methods: Callable) -> Callable:
    """Register __array_ufunc__."""
    def decorator(numpoly_function: Callable) -> Callable:
        """Register function."""
        for func in array_methods:
            assert func not in UFUNC_COLLECTION, (
                f"{func} already implemented")
            UFUNC_COLLECTION[func] = numpoly_function
        return numpoly_function
    return decorator


def implements(*array_functions: Callable) -> Callable:
    """Register __array_function__ and __array_ufunc__."""
    def decorator(numpoly_function: Callable) -> Callable:
        """Register function."""
        for func in array_functions:
            assert func not in FUNCTION_COLLECTION, (
                "{func} already implemented")
            FUNCTION_COLLECTION[func] = numpoly_function
            assert func not in UFUNC_COLLECTION, "{func} already implemented"
            UFUNC_COLLECTION[func] = numpoly_function
        return numpoly_function

    return decorator


def simple_dispatch(
        numpy_func: Callable,
        inputs: Sequence[Any],
        out: Optional[Tuple[ndpoly, ...]] = None,
        **kwargs: Any
) -> ndpoly:
    """
    Dispatch function between numpy and numpoly.

    Assumes that evaluation can be performed on the coefficients alone and that
    there are no change to the polynomials.

    Args:
        numpy_func:
            The numpy function to evaluate `inputs` on.
        inputs:
            One or more input arrays.
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        kwargs:
            Keyword args passed to `numpy_func`.

    Returns:
        Polynomial, where the coefficients from `input` are passed to
        `numpy_func` to create the output coefficients.

    """
    inputs = numpoly.align_polynomials(*inputs)
    keys = numpoly.aspolynomial(inputs[0] if out is None else out[0]).keys

    tmp = numpy_func(*[poly.values[keys[0]]for poly in inputs], **kwargs)
    if out is None:
        out_ = numpoly.ndpoly(
            exponents=inputs[0].exponents,
            shape=tmp.shape,
            names=inputs[0].indeterminants,
            dtype=tmp.dtype,
        )
    else:
        assert len(out) == 1
        out_ = out[0]
    out_.values[keys[0]] = tmp

    for key in keys[1:]:
        out_.values[key] = numpy_func(
            *[poly.values[key] for poly in inputs], **kwargs)

    if out is None:
        out_ = numpoly.clean_attributes(out_)
    return numpoly.aspolynomial(out_)



"""Split array into multiple sub-arrays along the 3rd axis (depth)."""
from __future__ import annotations
from typing import List

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.dsplit)
def dsplit(
    ary: PolyLike,
    indices_or_sections: numpy.typing.ArrayLike,
) -> List[ndpoly]:
    """
    Split array into multiple sub-arrays along the 3rd axis (depth).

    Please refer to the `split` documentation.  `dsplit` is equivalent
    to `split` with ``axis=2``, the array is always split along the third
    axis provided the array dimension is greater than or equal to 3.

    Examples:
        >>> poly = numpoly.monomial(8).reshape(2, 2, 2)
        >>> poly
        polynomial([[[1, q0],
                     [q0**2, q0**3]],
        <BLANKLINE>
                    [[q0**4, q0**5],
                     [q0**6, q0**7]]])
        >>> part1, part2 = numpoly.dsplit(poly, 2)
        >>> part1
        polynomial([[[1],
                     [q0**2]],
        <BLANKLINE>
                    [[q0**4],
                     [q0**6]]])
        >>> part2
        polynomial([[[q0],
                     [q0**3]],
        <BLANKLINE>
                    [[q0**5],
                     [q0**7]]])

    """
    ary = numpoly.aspolynomial(ary)
    results = numpy.dsplit(ary.values, indices_or_sections=indices_or_sections)
    return [
        numpoly.polynomial(
            result, names=ary.indeterminants, allocation=ary.allocation)
        for result in results
    ]



"""Load polynomial or pickled objects from ``.np{y,z}`` or pickled files."""
from __future__ import annotations
from typing import BinaryIO, Dict, Optional, Union
from os import PathLike

import numpy
import numpoly

from ..baseclass import ndpoly


def load(
        file: Union[BinaryIO, PathLike],
        mmap_mode: Optional[str] = None,
        allow_pickle: bool = False,
        fix_imports: bool = True,
        encoding: str = "ASCII",
) -> Union[ndpoly, Dict[str, ndpoly]]:
    """
    Load polynomial or pickled objects from ``.np{y,z}`` or pickled files.

    Args:
        file:
            The file to read. File-like objects must support the ``seek()``
            and ``read()`` methods. Pickled files require that the file-like
            object support the ``readline()`` method as well.
        mmap_mode:
            If not None, then memory-map the file, using the given mode (see
            `numpy.memmap` for a detailed description of the modes). A
            memory-mapped array is kept on disk. However, it can be accessed
            and sliced like any ndarray. Memory mapping is especially useful
            for accessing small fragments of large files without reading the
            entire file into memory.
        allow_pickle:
            Allow loading pickled object arrays stored in npy files. Reasons
            for disallowing pickles include security, as loading pickled data
            can execute arbitrary code. If pickles are disallowed, loading
            object arrays will fail.
        fix_imports:
            Only useful when loading Python 2 generated pickled files on
            Python 3, which includes npy/npz files containing object arrays.
            If `fix_imports` is True, pickle will try to map the old Python 2
            names to the new names used in Python 3.
        encoding:
            What encoding to use when reading Python 2 strings. Only useful
            when loading Python 2 generated pickled files in Python 3, which
            includes npy/npz files containing object arrays. Values other than
            'latin1', 'ASCII', and 'bytes' are not allowed, as they can corrupt
            numerical data.

    Returns:
        Data stored in the file. For ``.npz`` files, the returned dictionary
        class must be closed to avoid leaking file descriptors.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([q0, q1-1])
        >>> array = numpy.array([1, 2])
        >>> numpoly.savez("/tmp/savez.npz", a=array, p=poly)
        >>> numpoly.load("/tmp/savez.npz")
        {'a': array([1, 2]), 'p': polynomial([q0, q1-1])}

    """
    if isinstance(file, (str, bytes, PathLike)):
        with open(file, "rb") as src:
            return load(file=src, mmap_mode=mmap_mode,
                        allow_pickle=allow_pickle,
                        fix_imports=fix_imports, encoding=encoding)

    out = numpy.load(file=file, mmap_mode=mmap_mode, allow_pickle=allow_pickle,
                     fix_imports=fix_imports, encoding=encoding)
    if isinstance(out, numpy.lib.npyio.NpzFile):  # type: ignore
        out = dict(out)
        for key, value in list(out.items()):

            # Classical arrays untouched
            if not value.dtype.names:
                continue

            # Structured array are assumed polynomials
            del out[key]
            for name in value.dtype.names:
                if not name.isdigit():
                    length = len(name)
                    break
            key = key.split("-")
            names = tuple(key[:length])
            key = "-".join(key[length:])
            out[key] = numpoly.polynomial(value, names=names)

    elif out.dtype.names:
        names = numpy.load(file=file, mmap_mode=mmap_mode,
                           allow_pickle=allow_pickle,
                           fix_imports=fix_imports, encoding=encoding).tolist()
        out = numpoly.polynomial(out, names=names)
    return out



"""Evenly round to the given number of decimals."""
from __future__ import annotations
from typing import Optional

import numpy

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements, simple_dispatch


@implements(numpy.around, numpy.round)
def around(
    a: PolyLike,
    decimals: int = 0,
    out: Optional[ndpoly] = None,
) -> ndpoly:
    """
    Evenly round to the given number of decimals.

    Args:
        a:
            Input data.
        decimals:
            Number of decimal places to round to (default: 0). If decimals is
            negative, it specifies the number of positions to the left of the
            decimal point.
        out:
            Alternative output array in which to place the result. It must have
            the same shape as the expected output, but the type of the output
            values will be cast if necessary.

    Returns:
        An array of the same type as `a`, containing the rounded values. Unless
        `out` was specified, a new array is created.  A reference to the result
        is returned. The real and imaginary parts of complex numbers are
        rounded separately.  The result of rounding a float is a float.

    Examples:
        >>> q0 = numpoly.variable()
        >>> numpoly.around([0.37, 1.64*q0-2.45])
        polynomial([0.0, 2.0*q0-2.0])
        >>> numpoly.around([0.37, 1.64*q0-23.45], decimals=1)
        polynomial([0.4, 1.6*q0-23.4])
        >>> numpoly.around([0.37, 1.64*q0-23.45], decimals=-1)
        polynomial([0.0, -20.0])

    """
    return simple_dispatch(
        numpy_func=numpy.around,
        inputs=(a,),
        out=None if out is None else (out,),
        decimals=decimals,
    )



"""Polynomial functionality."""
from .call import call
from .decompose import decompose
from .derivative import derivative, gradient, hessian
from .isconstant import isconstant
from .lead_coefficient import lead_coefficient
from .lead_exponent import lead_exponent
from .divide import (
    get_division_candidate, poly_divide, poly_divmod, poly_remainder)
from .set_dimensions import set_dimensions
from .sortable_proxy import sortable_proxy
from .tonumpy import tonumpy

__all__ = (
    "call", "decompose", "derivative", "gradient", "hessian", "isconstant",
    "lead_coefficient", "lead_exponent", "poly_divide", "poly_divmod",
    "poly_remainder", "set_dimensions", "sortable_proxy", "tonumpy",
)



"""Return an array of zeros with the same shape and type as a given array."""
from __future__ import annotations
from typing import Any, Optional, Sequence

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements

Order = Any
try:
    from typing import Literal, Union
    Order = Union[None, Literal["C"], Literal["F"]]  # type: ignore
except ImportError:
    pass


@implements(numpy.zeros_like)
def zeros_like(
        a: PolyLike,
        dtype: Optional[numpy.typing.DTypeLike] = None,
        order: Order = None,
        subok: bool = True,
        shape: Optional[Sequence[int]] = None,
) -> ndpoly:
    """
    Return an array of zeros with the same shape and type as a given array.

    Args:
        a:
            The shape and data-type of `a` define these same attributes of
            the returned array.
        dtype:
            Overrides the data type of the result.
        order:
            Overrides the memory layout of the result. 'C' means C-order,
            'F' means F-order. If omitted: 'F' if `a` is Fortran contiguous,
            'C' otherwise.
        subok:
            If True, then the newly created array will use the sub-class
            type of 'a', otherwise it will be a base-class array. Defaults
            to True.
        shape:
            Overrides the shape of the result. If order='K' and the number of
            dimensions is unchanged, will try to keep order, otherwise,
            order='C' is implied.

    Returns:
        Array of zeros with the same shape and type as `a`.

    Examples:
        >>> poly = numpoly.monomial(3)
        >>> poly
        polynomial([1, q0, q0**2])
        >>> numpoly.zeros_like(poly)
        polynomial([0, 0, 0])

    """
    del subok
    a = numpoly.aspolynomial(a)
    return numpoly.polynomial(numpy.zeros_like(  # pylint: disable=unexpected-keyword-arg
        a.values[a.keys[0]], dtype=dtype, order=order, shape=shape))



"""Return element-wise remainder of division."""
from __future__ import annotations
from typing import Any, Optional

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements_ufunc

REMAINDER_ERROR_MSG = """\
Polynomial division differs from numerical division;
Use ``numpoly.poly_remainder`` to get polynomial remainder."""


@implements_ufunc(numpy.remainder)
def remainder(
    x1: PolyLike,
    x2: PolyLike,
    out: Optional[ndpoly] = None,
    where: numpy.typing.ArrayLike = True,
    **kwargs: Any,
) -> ndpoly:
    """
    Return element-wise remainder of numerical division.

    Args:
        x1:
            Dividend array.
        x2:
            Divisor array. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        The element-wise remainder of the quotient ``floor_divide(x1, x2)``.
        This is a scalar if both `x1` and `x2` are scalars.

    Examples:
        >>> numpoly.remainder([14, 7], 5)
        polynomial([4, 2])
        >>> numpoly.remainder(numpoly.variable(), 2)  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        numpoly.baseclass.FeatureNotSupported: Polynomial division ...
        Use ``numpoly.poly_remainder`` to get polynomial remainder.

    """
    x1, x2 = numpoly.align_polynomials(x1, x2)
    if not x1.isconstant() or not x2.isconstant():
        raise numpoly.FeatureNotSupported(REMAINDER_ERROR_MSG)
    where = None if where is None else numpy.asarray(where)
    return numpoly.polynomial(numpy.remainder(
        x1.tonumpy(), x2.tonumpy(), out=out, where=where, **kwargs))



"""Difference between consecutive elements of an array."""
from __future__ import annotations
from typing import Optional

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.ediff1d)
def ediff1d(
    ary: PolyLike,
    to_end: Optional[PolyLike] = None,
    to_begin: Optional[PolyLike] = None,
) -> ndpoly:
    """
    Difference between consecutive elements of an array.

    Args:
        ary:
            If necessary, will be flattened before the differences are taken.
        to_end:
            Polynomial(s) to append at the end of the returned differences.
        to_begin:
            Polynomial(s) to prepend at the beginning of the returned
            differences.

    Returns:
        The differences. Loosely, this is ``ary.flat[1:] - ary.flat[:-1]``.

    Examples:
        >>> poly = numpoly.monomial(4)
        >>> poly
        polynomial([1, q0, q0**2, q0**3])
        >>> numpoly.ediff1d(poly)
        polynomial([q0-1, q0**2-q0, q0**3-q0**2])
        >>> q0, q1 = numpoly.variable(2)
        >>> numpoly.ediff1d(poly, to_begin=q0, to_end=[1, q1])
        polynomial([q0, q0-1, q0**2-q0, q0**3-q0**2, 1, q1])

    """
    ary = numpoly.aspolynomial(ary).ravel()
    arys_ = [ary[1:]-ary[:-1]]
    if to_end is not None:
        arys_.append(numpoly.aspolynomial(to_end).ravel())
    if to_begin is not None:
        arys_.insert(0, numpoly.aspolynomial(to_begin).ravel())
    arys = tuple(numpoly.aspolynomial(ary) for ary in arys_)
    if len(arys) > 1:
        arys = numpoly.align_dtype(*arys)
        arys = numpoly.align_exponents(*arys)
        arys = numpoly.align_indeterminants(*arys)

    out = numpoly.ndpoly(
        exponents=arys[0].exponents,
        shape=(sum([ary.size for ary in arys]),),
        names=arys[0].names,
        dtype=ary[0].dtype,
    )

    idx = 0
    for ary in arys:
        for key in ary.keys:
            out.values[key][idx:idx+ary.size] = ary.values[key]
        idx += ary.size

    return out



"""Compute the outer product of two vectors."""
from __future__ import annotations
from typing import Optional

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.outer)
def outer(
    a: PolyLike,
    b: PolyLike,
    out: Optional[ndpoly] = None,
) -> ndpoly:
    """
    Compute the outer product of two vectors.

    Given two vectors, ``a = [a0, a1, ..., aM]`` and
    ``b = [b0, b1, ..., bN]``, the outer product is::

        [[a0*b0  a0*b1 ... a0*bN ]
         [a1*b0    .
         [ ...          .
         [aM*b0            aM*bN ]]

    Args:
        a:
            First input vector. Input is flattened if not already
            1-dimensional.
        b:
            Second input vector. Input is flattened if not already
            1-dimensional.
        out:
            A location where the result is stored.

    Returns:
        ``out[i, j] = a[i] * b[j]``

    Examples:
        >>> poly = numpoly.variable(3)
        >>> const = numpy.arange(5)
        >>> numpoly.outer(poly, const)
        polynomial([[0, q0, 2*q0, 3*q0, 4*q0],
                    [0, q1, 2*q1, 3*q1, 4*q1],
                    [0, q2, 2*q2, 3*q2, 4*q2]])

    """
    a, b = numpoly.align_exponents(a, b)
    a = a.ravel()[:, numpy.newaxis]
    b = b.ravel()[numpy.newaxis, :]
    return numpoly.multiply(a, b, out=out)



"""Broadcast any number of arrays against each other."""
from __future__ import annotations
from typing import Any, List

import numpy
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.broadcast_arrays)
def broadcast_arrays(*args: PolyLike, **kwargs: Any) -> List[ndpoly]:
    """
    Broadcast any number of arrays against each other.

    Args:
        args:
            The arrays to broadcast.
        subok:
            If True, then sub-classes will be passed-through, otherwise the
            returned arrays will be forced to be a base-class array (default).

    Returns:
        These arrays are views on the original arrays.  They are typically not
        contiguous.  Furthermore, more than one element of a broadcasted array
        may refer to a single memory location. If you need to write to the
        arrays, make copies first. While you can set the ``writable`` flag
        True, writing to a single output value may end up changing more than
        one location in the output array.

    Examples:
        >>> poly1 = numpoly.monomial(3).reshape(1, 3)
        >>> poly1
        polynomial([[1, q0, q0**2]])
        >>> poly2 = numpoly.monomial(2).reshape(2, 1)
        >>> poly2
        polynomial([[1],
                    [q0]])
        >>> res1, res2 = numpoly.broadcast_arrays(poly1, poly2)
        >>> res1
        polynomial([[1, q0, q0**2],
                    [1, q0, q0**2]])
        >>> res2
        polynomial([[1, 1, 1],
                    [q0, q0, q0]])

    """
    args_ = [numpoly.aspolynomial(arg) for arg in args]
    results = numpy.broadcast_arrays(*[arg.values for arg in args_], **kwargs)
    return [numpoly.aspolynomial(result, names=arg.indeterminants)
            for result, arg in zip(results, args_)]



"""Return element-wise remainder of polynomial division."""
from __future__ import annotations
from typing import Any, Optional

import numpy
import numpy.typing

from ...baseclass import ndpoly, PolyLike
from ...dispatch import implements_function
from .divmod import poly_divmod


@implements_function(numpy.remainder)
def poly_remainder(
        x1: PolyLike,
        x2: PolyLike,
        out: Optional[ndpoly] = None,
        where: numpy.typing.ArrayLike = True,
        **kwargs: Any,
) -> ndpoly:
    """
    Return element-wise remainder of polynomial division.

    Args:
        x1:
            Dividend array.
        x2:
            Divisor array. If ``x1.shape != x2.shape``, they must be
            broadcastable to a common shape (which becomes the shape of the
            output).
        out:
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or
            `None`, a freshly-allocated array is returned. A tuple (possible
            only as a keyword argument) must have length equal to the number of
            outputs.
        where:
            This condition is broadcast over the input. At locations where the
            condition is True, the `out` array will be set to the ufunc result.
            Elsewhere, the `out` array will retain its original value. Note
            that if an uninitialized `out` array is created via the default
            ``out=None``, locations within it where the condition is False will
            remain uninitialized.
        kwargs:
            Keyword args passed to numpy.ufunc.

    Returns:
        The element-wise remainder of the quotient ``floor_divide(x1, x2)``.
        This is a scalar if both `x1` and `x2` are scalars.

    Notes:
        Unlike numbers, this returns the polynomial division and polynomial
        remainder. This means that this function is _not_ backwards compatible
        with ``numpy.remainder`` for constants. For example:
        ``numpoly.remainder(11, 2) == 1`` while
        ``numpoly.poly_remainder(11, 2) == 0``.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> denominator = [q0*q1**2+2*q0**3*q1**2, -2+q0*q1**2]
        >>> numerator = -2+q0*q1**2
        >>> numpoly.poly_remainder(denominator, numerator)
        polynomial([4.0*q0**2+2.0, 0.0])

    """
    dividend, remainder = poly_divmod(x1, x2, out=out, where=where, **kwargs)
    return remainder



"""Return a new array of given shape and type, filled with `fill_value`."""
from __future__ import annotations
from typing import Optional, Sequence, Union

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.full)
def full(
    shape: Union[int, Sequence[int]],
    fill_value: PolyLike,
    dtype: Optional[numpy.typing.DTypeLike] = None,
    order: str = "C",
) -> ndpoly:
    """
    Return a new array of given shape and type, filled with `fill_value`.

    Args:
        shape:
            Shape of the new array, e.g., ``(2, 3)`` or ``2``.
        fill_value:
            Fill value. Must be broadcast compatible with `shape`.
        dtype:
            The desired data-type for the array  The default, None, means
            inherit from `fill_value`.
        order:
            Whether to store multidimensional data in C- or Fortran-contiguous
            (row- or column-wise) order in memory. Valid values: "C", "F".

    Returns:
        Array of `fill_value` with the given shape, dtype, and order.

    Examples:
        >>> numpoly.full((2, 4), 4)
        polynomial([[4, 4, 4, 4],
                    [4, 4, 4, 4]])
        >>> q0 = numpoly.variable()
        >>> numpoly.full(3, q0**2-1)
        polynomial([q0**2-1, q0**2-1, q0**2-1])

    """
    fill_value = numpoly.aspolynomial(fill_value)
    if dtype is None:
        dtype = fill_value.dtype
    shape = tuple((shape,) if isinstance(shape, int) else shape)
    out = numpoly.ndpoly(
        exponents=fill_value.exponents,
        shape=shape,
        names=fill_value.indeterminants,
        dtype=dtype,
        order=order,
    )
    for key in fill_value.keys:
        out.values[key] = fill_value.values[key]
    return out



"""Construct an array by repeating A the number of times given by reps."""
from __future__ import annotations

import numpy
import numpy.typing
import numpoly

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements


@implements(numpy.tile)
def tile(A: PolyLike, reps: numpy.typing.ArrayLike) -> ndpoly:
    """
    Construct an array by repeating A the number of times given by reps.

    If `reps` has length ``d``, the result will have dimension of
    ``max(d, A.ndim)``.

    If ``A.ndim < d``, `A` is promoted to be d-dimensional by prepending new
    axes. So a shape (3,) array is promoted to (1, 3) for 2-D replication,
    or shape (1, 1, 3) for 3-D replication. If this is not the desired
    behavior, promote `A` to d-dimensions manually before calling this
    function.

    If ``A.ndim > d``, `reps` is promoted to `A`.ndim by pre-pending 1's to it.
    Thus for an `A` of shape (2, 3, 4, 5), a `reps` of (2, 2) is treated as
    (1, 1, 2, 2).

    Args:
        A:
            The input array.
        reps:
            The number of repetitions of `A` along each axis.

    Returns:
        The tiled output array.

    Examples:
        >>> q0 = numpoly.variable()
        >>> numpoly.tile(q0, 4)
        polynomial([q0, q0, q0, q0])
        >>> poly = numpoly.polynomial([[1, q0-1], [q0**2, q0]])
        >>> numpoly.tile(poly, 2)
        polynomial([[1, q0-1, 1, q0-1],
                    [q0**2, q0, q0**2, q0]])
        >>> numpoly.tile(poly, [2, 1])
        polynomial([[1, q0-1],
                    [q0**2, q0],
                    [1, q0-1],
                    [q0**2, q0]])

    """
    A = numpoly.aspolynomial(A)
    result = numpy.tile(A.values, reps=reps)
    return numpoly.aspolynomial(result, names=A.indeterminants)



"""Collection of numpy wrapper functions."""
from .absolute import absolute as abs, absolute
from .add import add
from .amax import amax, amax as max
from .amin import amin, amin as min
from .any import any
from .all import all
from .allclose import allclose
from .apply_along_axis import apply_along_axis
from .apply_over_axes import apply_over_axes
from .argmax import argmax
from .argmin import argmin
from .around import around, around as round, around as round_
from .array_repr import array_repr
from .array_split import array_split
from .array_str import array_str
from .atleast_1d import atleast_1d
from .atleast_2d import atleast_2d
from .atleast_3d import atleast_3d
from .broadcast_arrays import broadcast_arrays
from .ceil import ceil
from .choose import choose
from .common_type import common_type
from .concatenate import concatenate
from .copyto import copyto
from .count_nonzero import count_nonzero
from .cumsum import cumsum
from .det import det
from .diag import diag
from .diff import diff
from .diagonal import diagonal
from .true_divide import true_divide, true_divide as divide
from .divmod import divmod
from .dsplit import dsplit
from .dstack import dstack
from .ediff1d import ediff1d
from .equal import equal
from .expand_dims import expand_dims
from .floor import floor
from .floor_divide import floor_divide
from .full import full
from .full_like import full_like
from .greater import greater
from .greater_equal import greater_equal
from .hstack import hstack
from .hsplit import hsplit
from .less import less
from .less_equal import less_equal
from .load import load
from .inner import inner
from .isclose import isclose
from .isfinite import isfinite
from .loadtxt import loadtxt
from .logical_and import logical_and
from .logical_or import logical_or
from .matmul import matmul
from .maximum import maximum
from .minimum import minimum
from .mean import mean
from .moveaxis import moveaxis
from .multiply import multiply
from .negative import negative
from .nonzero import nonzero
from .not_equal import not_equal
from .ones import ones
from .ones_like import ones_like
from .outer import outer
from .positive import positive
from .power import power
from .prod import prod
from .remainder import remainder as mod, remainder
from .repeat import repeat
from .reshape import reshape
from .result_type import result_type
from .rint import rint
from .roots import roots
from .save import save
from .savetxt import savetxt
from .savez import savez
from .savez_compressed import savez_compressed
from .split import split
from .square import square
from .stack import stack
from .subtract import subtract
from .sum import sum
from .tile import tile
from .transpose import transpose
from .vsplit import vsplit
from .vstack import vstack
from .where import where
from .zeros import zeros
from .zeros_like import zeros_like



"""Save polynomial array to a binary file in NumPy ``.npy`` format."""
from __future__ import annotations
from os import PathLike

import numpy
import numpoly

from ..baseclass import PolyLike
from ..dispatch import implements


@implements(numpy.save)
def save(
        file: PathLike,
        arr: PolyLike,
        allow_pickle: bool = True,
        fix_imports: bool = True,
) -> None:
    """
    Save polynomial array to a binary file in NumPy ``.npy`` format.

    Args:
        file:
            File or filename to which the data is saved. If file is a
            file-object, then the filename is unchanged. If file is a string
            or Path, a ``.npy`` extension will be appended to the filename if
            it does not already have one.
        arr:
            Array data to be saved.
        allow_pickle:
            Allow saving object arrays using Python pickles. Reasons for
            disallowing pickles include security (loading pickled data can
            execute arbitrary code) and portability (pickled objects may not be
            loadable on different Python installations, for example if the
            stored objects require libraries that are not available, and not
            all pickled data is compatible between Python 2 and Python 3).
        fix_imports:
            Only useful in forcing objects in object arrays on Python 3 to be
            pickled in a Python 2 compatible way. If `fix_imports` is True,
            pickle will try to map the new Python 3 names to the old module
            names used in Python 2, so that the pickle data stream is readable
            with Python 2.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([1, q0, q1**2-1])
        >>> numpoly.save("/tmp/example1.npy", poly)
        >>> numpoly.load("/tmp/example1.npy")
        polynomial([1, q0, q1**2-1])

    """
    if isinstance(file, (str, bytes, PathLike)):
        with open(file, "wb") as src:
            return save(src, arr=arr,
                        allow_pickle=allow_pickle, fix_imports=fix_imports)
    arr = numpoly.aspolynomial(arr)
    numpy.save(file=file, arr=arr.values,
               allow_pickle=allow_pickle, fix_imports=fix_imports)
    numpy.save(file=file, arr=numpy.array(arr.names),
               allow_pickle=allow_pickle, fix_imports=fix_imports)



"""Align polynomials."""
from __future__ import annotations
from typing import Tuple

import numpy
import numpoly
from .baseclass import ndpoly, PolyLike


def align_polynomials(*polys: PolyLike) -> Tuple[ndpoly, ...]:
    """
    Align polynomial such that dimensionality, shape, etc. are compatible.

    Alignment includes shape (for broadcasting), indeterminants, exponents and
    dtype.

    Args:
        polys:
            Polynomial to make adjustment to.

    Returns:
        Same as ``polys``, but internal adjustments made to make them
        compatible for further operations.

    Examples:
        >>> q0 = numpoly.variable()
        >>> q0q1 = numpoly.variable(2)
        >>> q0
        polynomial(q0)
        >>> q0.coefficients
        [1]
        >>> q0.indeterminants
        polynomial([q0])
        >>> q0, _ = numpoly.align_polynomials(q0, q0q1.astype(float))
        >>> q0
        polynomial([q0, q0])
        >>> q0.coefficients
        [array([1., 1.]), array([0., 0.])]
        >>> q0.indeterminants
        polynomial([q0, q1])

    """
    # polys = numpoly.broadcast_arrays(*polys)
    polys = align_shape(*polys)
    polys = align_indeterminants(*polys)
    polys = align_exponents(*polys)
    polys = align_dtype(*polys)
    return polys


def align_shape(*polys: PolyLike) -> Tuple[ndpoly, ...]:
    """
    Align polynomial by shape.

    Args:
        polys:
            Polynomial to make adjustment to.

    Returns:
        Same as ``polys``, but internal adjustments made to make them
        compatible for further operations.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly1 = 4*q0
        >>> poly2 = numpoly.polynomial([[2*q0+1, 3*q0-q1]])
        >>> poly1.shape
        ()
        >>> poly2.shape
        (1, 2)
        >>> poly1, poly2 = numpoly.align_shape(poly1, poly2)
        >>> poly1
        polynomial([[4*q0, 4*q0]])
        >>> poly2
        polynomial([[2*q0+1, -q1+3*q0]])
        >>> poly1.shape
        (1, 2)
        >>> poly2.shape
        (1, 2)

    """
    # return tuple(numpoly.broadcast_arrays(*polys))
    polys_ = tuple(numpoly.aspolynomial(poly) for poly in polys)
    common = numpy.ones((), dtype=int)
    for poly in polys_:
        if poly.size:
            common = numpy.ones(poly.coefficients[0].shape, dtype=int)*common

    polys_ = tuple(poly.from_attributes(
        exponents=poly.exponents,
        coefficients=tuple(coeff*common for coeff in poly.coefficients),
        names=poly.indeterminants,
    ) for poly in polys_)
    assert numpy.all([common.shape == poly.shape for poly in polys_])
    return tuple(polys_)


def align_indeterminants(*polys: PolyLike) -> Tuple[ndpoly, ...]:
    """
    Align polynomial by indeterminants.

    Args:
        polys:
            Polynomial to make adjustment to.

    Returns:
        Same as ``polys``, but internal adjustments made to make them
        compatible for further operations.

    Examples:
        >>> q0 = numpoly.variable()
        >>> poly1 = numpoly.polynomial(2*q0+1)
        >>> q0, q1 = numpoly.variable(2)
        >>> poly2 = numpoly.polynomial(3*q0-q1)
        >>> poly1.indeterminants
        polynomial([q0])
        >>> poly2.indeterminants
        polynomial([q0, q1])
        >>> poly1, poly2 = numpoly.align_indeterminants(poly1, poly2)
        >>> poly1
        polynomial(2*q0+1)
        >>> poly2
        polynomial(-q1+3*q0)
        >>> poly1.indeterminants
        polynomial([q0, q1])
        >>> poly2.indeterminants
        polynomial([q0, q1])

    """
    polys_ = [numpoly.aspolynomial(poly) for poly in polys]
    common_names = tuple(sorted({
        str(name) for poly in polys_ for name in poly.names}))
    if not common_names:
        return tuple(polys_)

    for idx, poly in enumerate(polys_):
        indices = numpy.array([
            common_names.index(name)
            for name in poly.names
            if name in common_names
        ])
        exponents = numpy.zeros(
            (len(poly.keys), len(common_names)), dtype=int)
        if indices.size:
            exponents[:, indices] = poly.exponents
        polys_[idx] = numpoly.ndpoly.from_attributes(
            exponents=exponents,
            coefficients=poly.coefficients,
            names=common_names,
            retain_coefficients=True,
            retain_names=True,
        )
    assert all([polys_[0].names == poly.names for poly in polys_])
    return tuple(polys_)


def align_exponents(*polys: PolyLike) -> Tuple[ndpoly, ...]:
    """
    Align polynomials such that the exponents are the same.

    Aligning exponents assumes that the indeterminants is also aligned.

    Args:
        polys:
            Polynomial to make adjustment to.

    Returns:
        Same as ``polys``, but internal adjustments made to make them
        compatible for further operations.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly1 = q0*q1
        >>> poly2 = numpoly.polynomial([q0**5, q1**3-1])
        >>> poly1.exponents
        array([[1, 1]], dtype=uint32)
        >>> poly2.exponents
        array([[0, 0],
               [0, 3],
               [5, 0]], dtype=uint32)
        >>> poly1, poly2 = numpoly.align_exponents(poly1, poly2)
        >>> poly1
        polynomial(q0*q1)
        >>> poly2
        polynomial([q0**5, q1**3-1])
        >>> poly1.exponents
        array([[1, 1],
               [0, 0],
               [0, 3],
               [5, 0]], dtype=uint32)
        >>> poly2.exponents
        array([[1, 1],
               [0, 0],
               [0, 3],
               [5, 0]], dtype=uint32)

    """
    polys_ = [numpoly.aspolynomial(poly) for poly in polys]
    if not all(
            polys_[0].names == poly.names
            for poly in polys_
    ):
        polys_ = list(align_indeterminants(*polys_))

    global_exponents = [tuple(exponent) for exponent in polys_[0].exponents]

    for poly in polys_[1:]:
        global_exponents.extend([tuple(exponent)
                                 for exponent in poly.exponents
                                 if tuple(exponent) not in global_exponents])

    for idx, poly in enumerate(polys_):
        lookup = {
            tuple(exponent): coefficient
            for exponent, coefficient in zip(
                poly.exponents, poly.coefficients)
        }

        zeros = numpy.zeros(poly.shape, dtype=poly.dtype)
        coefficients = [lookup.get(exponent, zeros)
                        for exponent in global_exponents]
        polys_[idx] = poly.from_attributes(
            exponents=global_exponents,
            coefficients=coefficients,
            names=poly.names,
            retain_coefficients=True,
            retain_names=True,
        )
    return tuple(polys_)


def align_dtype(*polys: PolyLike) -> Tuple[ndpoly, ...]:
    """
    Align polynomial by shape.

    Args:
        polys (numpoly.ndpoly):
            Polynomial to make adjustment to.

    Returns:
        (Tuple[numpoly.ndpoly, ...]):
            Same as ``polys``, but internal adjustments made to make them
            compatible for further operations.

    Examples:
        >>> q0 = numpoly.variable()
        >>> q0.dtype.name
        'int64'
        >>> poly, _ = numpoly.align_dtype(q0, 4.5)
        >>> poly.dtype.name
        'float64'

    """
    polys_ = [numpoly.aspolynomial(poly) for poly in polys]
    dtype = numpy.asarray(numpy.sum(numpy.array([
        numpy.array(True, dtype=poly.dtype) for poly in polys_]))).dtype
    polys_ = [numpoly.ndpoly.from_attributes(
        exponents=poly.exponents,
        coefficients=poly.coefficients,
        names=poly.indeterminants,
        dtype=dtype,
        retain_coefficients=True,
        retain_names=True,
    ) for poly in polys_]
    return tuple(polys_)



"""Return the cumulative sum of the elements along a given axis."""
from __future__ import annotations
from typing import Optional

import numpy
import numpy.typing

from ..baseclass import ndpoly, PolyLike
from ..dispatch import implements, simple_dispatch


@implements(numpy.cumsum)
def cumsum(
    a: PolyLike,
    axis: Optional[int] = None,
    dtype: Optional[numpy.typing.DTypeLike] = None,
    out: Optional[ndpoly] = None,
) -> ndpoly:
    """
    Return the cumulative sum of the elements along a given axis.

    Args:
        a:
            Input array.
        axis:
            Axis along which the cumulative sum is computed. The default
            (None) is to compute the cumsum over the flattened array.
        dtype:
            Type of the returned array and of the accumulator in which the
            elements are summed.  If `dtype` is not specified, it defaults
            to the dtype of `a`, unless `a` has an integer dtype with a
            precision less than that of the default platform integer.  In
            that case, the default platform integer is used.
        out:
            Alternative output array in which to place the result. It must
            have the same shape and buffer length as the expected output
            but the type will be cast if necessary.

    Args:
        A new array holding the result is returned unless `out` is specified,
        in which case a reference to `out` is returned. The result has the same
        size as `a`, and the same shape as `a` if `axis` is not None or `a` is
        a 1-d array.

    Examples:
        >>> q0, q1 = numpoly.variable(2)
        >>> poly = numpoly.polynomial([[1, q0, 3], [4, 5, q1]])
        >>> poly
        polynomial([[1, q0, 3],
                    [4, 5, q1]])
        >>> numpoly.cumsum(poly)
        polynomial([1, q0+1, q0+4, q0+8, q0+13, q1+q0+13])
        >>> numpoly.cumsum(poly, dtype=float, axis=1)
        polynomial([[1.0, q0+1.0, q0+4.0],
                    [4.0, 9.0, q1+9.0]])
        >>> numpoly.cumsum(poly, axis=0)
        polynomial([[1, q0, 3],
                    [5, q0+5, q1+3]])

    """
    return simple_dispatch(
        numpy_func=numpy.cumsum,
        inputs=(a,),
        out=None if out is None else (out,),
        axis=axis,
        dtype=dtype,
    )
